/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define VERSION_REV 28
#define VERSION_DATE "2017-05-18"

#include "flxoctave.h"

#if defined(_MSC_VER)
  #include "flxliboctave_export.h"
#else
  #define FLXLIBOCTAVE_EXPORT
#endif

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlib")
#endif

flxlibs_interface flx_iface;


#include <octave/config.h>
#include <octave/oct.h>
#include <octave/octave.h>
#include <octave/parse.h>
#include <octave/toplev.h> /* do_octave_atexit */

#if FLX_DEBUG
  int FlxOctaveBox::Cinst = 0;
#endif
FlxOctaveBox OctaveBox;
FlxData* flxdata_ref = NULL;
  

extern "C" FLXLIBOCTAVE_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  Octave-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("octave_addpath", new FlxObjReadOctave_addpath()); 
    flx_iface.add_obj("octave_mtx", new FlxObjReadOctaveMtx()); 
    flx_iface.add_obj("octave_cmd", new FlxObjReadOctave_cmd());
    flx_iface.add_obj("octave_setvar", new FlxObjReadOctave_setvar()); 
  // functions
    flxdata_ref = dataBox;
    flx_iface.add_fun( "octave_lsf",new FunReadFunOctaveLSF() );
    flx_iface.add_fun( "octave_fun",new FunReadFunOctaveFun() );
    flx_iface.add_fun( "octave_getvar",new FunReadFunOctaveGetVar() );
}

extern "C" FLXLIBOCTAVE_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}


void FlxObjOctaveMtx::task()
{
  // LINKs: 
  //  https://mailman.cae.wisc.edu/pipermail/help-octave/2009-April/034429.html
  //  http://wiki.octave.org/Tips_and_tricks#C.2B.2B
  
  // get the vector
    tuint N = 0;
    tdouble* vp = data->ConstMtxBox.get_Vec(inpVecName->eval(),N);
  // convert to Octave format
    Matrix rvVec(N,1);
    for (tuint i=0;i<N;++i) {
      rvVec(i,0) = vp[i];
    }
    octave_value_list functionArguments;
    functionArguments(0) = rvVec;
  // return result
    const octave_value_list result = feval(funname,functionArguments, 1);
    Matrix rM = result(0).matrix_value();
    tuint r = rM.rows();
    tuint c = rM.columns();
    tdouble* dp = data->ConstMtxBox.get_Mtx(resVecName->eval_word(true),r,c);
    size_t count = 0;
    for (tuint i=0;i<r;++i) {
      for (tuint j=0;j<c;++j) {
	dp[count] = rM(i,j);
	++count;
      }
    }
}

FlxObjBase* FlxObjReadOctaveMtx::read()
{
  FlxString* resVecName = new FlxString(false,false);
  FlxMtxConstFun* inpVecName = NULL;
  try {
    reader->getChar('=');
    // obtain the name of the Octave function to call
      if ( reader->getNextType() != ReadStream::STRING ) {
	std::ostringstream ssV;
	ssV << "Name of the Octave function to call expected.";
	throw FlxException_NeglectInInteractive("FlxObjReadOctaveMtx::read_1", ssV.str(),reader->getCurrentPos());
      }
      const std::string fname = reader->getWord(false,false);
    reader->getChar('(');
    inpVecName = new FlxMtxConstFun(true);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjOctaveMtx(get_doLog(),fname,resVecName,inpVecName);
  } catch (FlxException& e) {
    delete resVecName;
    if (inpVecName) delete inpVecName;
    throw;
  }
}


FunOctaveLSF::FunOctaveLSF(const std::string& funname, RBRV_constructor* RndBox, std::string rbrvsetn, std::vector<FunBase*> *ParaListV)
: FunBaseFun_multPara(ParaListV), funname(funname), RndBox(RndBox), NOX(RndBox->get_NOX()), rvv(NOX), rbrvsetn(rbrvsetn)
{
  OctaveBox.init_octave();  
}

FunOctaveLSF::~FunOctaveLSF()
{
  delete RndBox;
}

const tdouble FunOctaveLSF::calc()
{
  // LINKs: 
  //  https://mailman.cae.wisc.edu/pipermail/help-octave/2009-April/034429.html
  //  http://wiki.octave.org/Tips_and_tricks#C.2B.2B
  
  Matrix rvVec(NOX,1);
  RndBox->get_x_Vec(rvv.get_tmp_vptr());
  for (size_t i=0;i<NOX;++i) {
    rvVec(i,0) = rvv[i];
  }
  
  octave_value_list functionArguments;
  functionArguments(0) = rvVec;
  if (ParaList->size()>0) {
    for (size_t i = 0; i < ParaList->size(); ++i) {
      functionArguments(i+1) = ParaList->operator[](i)->calc();
    }
  }
  
  const octave_value_list result = feval(funname,functionArguments, 1);
  
  return result(0).scalar_value ();
}

const std::string FunOctaveLSF::write()
{
  std::string str1 = write_v() + '(' + funname;
  str1 += '(' + rbrvsetn + ')';
  if (ParaList->size()>0) {
    for (size_t i = 0; i < ParaList->size(); ++i) {
      str1 += ',' + ParaList->operator[](i)->write();
    }
  }
  str1 += ')';
  return str1;
}

const std::string FunOctaveLSF::write_v()
{
  return "octave_lsf";
}


FunBase* FunReadFunOctaveLSF::read( bool errSerious )
{
  // obtain the name of the Octave function to call
    if ( reader->getNextType() != ReadStream::STRING ) {
      std::ostringstream ssV;
      ssV << "Name of the Octave function to call expected.";
      FlxError(errSerious,"FunReadFunOctaveLSF::read_1", ssV.str(), reader->getCurrentPos());
    }
    const std::string fname = reader->getWord(false,errSerious);
    // get the rbrv-set
      reader->getChar(',',errSerious);
      FlxString st(false,errSerious);
      std::string rbrvsetn = st.eval(true);
    // get relevant variables
      std::vector<RBRV_set_base*> setvec;
      RBRV_constructor::find_dependent_sets(rbrvsetn,setvec,data->rbrv_box);
      RBRV_constructor* RndBox = new RBRV_constructor(setvec);
  try {
    
  // obtain additional paramters
    if (reader->whatIsNextChar()==',') reader->getChar(',',errSerious);

  return new FunOctaveLSF(fname, RndBox, rbrvsetn, read_parameters(-1,errSerious) ); 
  
  } catch (FlxException &e) {
    FLXMSG("FunReadFunOctaveLSF::read_2",1);
    delete RndBox;
    throw;
  }
}


// ------------------------------------------------------------------------------------------------


FunOctaveFun::FunOctaveFun(const std::string& funname, FlxMtxConstFun* inpVecF, FlxObjBase* loopblock, std::vector< FunBase* >* ParaListV)
: FunBaseFun_multPara(ParaListV), funname(funname), inpVecF(inpVecF), loopblock(loopblock)
{
  OctaveBox.init_octave();  
}

FunOctaveFun::~FunOctaveFun()
{
  if (loopblock) delete loopblock;
  delete inpVecF;
}

const tdouble FunOctaveFun::calc()
{
  // run the code-block
    if (loopblock) loopblock->exec();
  // get the matrix
    const std::string ivn = inpVecF->eval();
    tuint Nr = 0;
    tuint Nc = 0;
    tdouble* dvp = data->ConstMtxBox.get_Mtx(ivn,Nr,Nc,false);
    // convert to Octave format
      Matrix rvMtx(Nr,Nc);
      for (tuint i=0;i<Nr;++i) {
	for (tuint j=0;j<Nc;++j) {
	rvMtx(i,j) = *dvp;
	++dvp;
	}
      }
    octave_value_list functionArguments;
    functionArguments(0) = rvMtx;
    if (ParaList->size()>0) {
      for (size_t i = 0; i < ParaList->size(); ++i) {
	functionArguments(i+1) = ParaList->operator[](i)->calc();
      }
    }
  // return result
    const octave_value_list result = feval(funname,functionArguments, 1);
    if (result(0).is_scalar_type()==false) {
      throw FlxException("FunOctaveFun::calc", "Return type not supported.");
    }
    return result(0).scalar_value();
}

const std::string FunOctaveFun::write_v()
{
  return "octave_fun";
}

const std::string FunOctaveFun::write()
{
  std::string str1 = write_v() + '(' + funname + ',' + inpVecF->write() + ",{...}";
  if (ParaList->size()>0) {
    for (size_t i = 0; i < ParaList->size(); ++i) {
      str1 += ',' + ParaList->operator[](i)->write();
    }
  }
  str1 += ')';
  return str1;
}

const bool FunOctaveFun::dependOn_Const(const tdouble*const thenumber)
{
  throw FlxException_NotImplemented("FunOctaveFun::dependOn_Const");
}

FunBase* FunReadFunOctaveFun::read(bool errSerious)
{
  // obtain the name of the Octave function to call
    if ( reader->getNextType() != ReadStream::STRING ) {
      std::ostringstream ssV;
      ssV << "Name of the Octave function to call expected.";
      FlxError(errSerious,"FunReadFunOctaveFun::read_1", ssV.str(), reader->getCurrentPos());
    }
    const std::string fname = reader->getWord(false,errSerious);
  // get name of input MtxConst
    reader->getChar(',');
   FlxMtxConstFun* inputVecF = new FlxMtxConstFun(true);
   FlxObjBase* loopblock = NULL;
  try {
    if (reader->whatIsNextChar()==',') {
    // the code to execute beforehand
      reader->getChar(',');
      loopblock = FlxObjReadCodeBlock::read_block(true,errSerious);
    // obtain additional paramters
      if (reader->whatIsNextChar()==',') reader->getChar(',',errSerious);
    }
    return new FunOctaveFun(fname, inputVecF, loopblock, read_parameters(-1,errSerious) );
  } catch (FlxException &e) {
    FLXMSG("FunReadFunOctaveFun::read_3",1);
    if (loopblock) delete loopblock;
    delete inputVecF;
    throw;
  }
}


// ------------------------------------------------------------------------------------------------


FlxOctave_addpath::FlxOctave_addpath(const bool dolog, FlxString* Path)
: FlxObjBase(dolog), Path(Path)
{
   OctaveBox.init_octave();
}

void FlxOctave_addpath::task()
{
  const std::string psn = Path->eval(false);
  octave_value_list functionArguments;
  functionArguments(0) = psn;
  const octave_value_list result = feval("addpath",functionArguments, 0);
}

FlxObjBase* FlxObjReadOctave_addpath::read()
{
  reader->getChar('(');
  FlxString* ps = new FlxString(false,false);
  try {
    reader->getChar(')');
    return new FlxOctave_addpath(get_doLog(),ps);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadOctave_addpath::read",1);
    delete ps;
    throw;
  }
}


// ------------------------------------------------------------------------------------------------
  
FlxOctaveBox::FlxOctaveBox()
: initialized(false)
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxOctaveBox' created ...";
      throw FlxException("FlxOctaveBox::FlxOctaveBox", ssV.str() );
    }
  #endif
}

FlxOctaveBox::~FlxOctaveBox()
{
  if (initialized) {
    //do_octave_atexit();
    clean_up_and_exit(0);
    initialized = false;
  }
}

void FlxOctaveBox::init_octave_actual()
{
  if (initialized) return;
  // Initialize Octave
    string_vector octave_argv(2);
    octave_argv(0) = "embedded";
    octave_argv(1) = "-q";
    if (!octave_main(2, octave_argv.c_str_vec(), 1)) {
      throw FlxException("FlxOctaveBox::init_octave_actual","Octave interpreter initialization failed");
    }
//     const char * argvv [] = {"" /* name of program, not relevant */,  "--silent"};
//     const char * argvv [] = {"" /* name of program, not relevant */,  "--silent"};
//     octave_main (2, (char **) argvv, true /* embedded */);

    initialized = true;
    
}

// ------------------------------------------------------------------------------------------------


void FlxOctave_cmd::task()
{
  const std::string psn = cmd->eval(false);
  int parse_status;
  octave_value v = eval_string (psn, false, parse_status);
}

FlxObjBase* FlxObjReadOctave_cmd::read()
{
  reader->getChar('(');
  FlxString* ps = new FlxString(true,false);
  try {
    reader->getChar(')');
    return new FlxOctave_cmd(get_doLog(),ps);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadOctave_cmd::read",1);
    delete ps;
    throw;
  }
}


// ------------------------------------------------------------------------------------------------

void FlxOctave_setvar::task()
{
  #if FLX_DEBUG
    if (sfun && mfun) throw FlxException_Crude("FlxOctave_setvar::task");
  #endif
  // evaluate the name of the variable
    const std::string vn = vname->eval_word(false);
  if (sfun) {
    tdouble dv = sfun->calc();
    octave_value val = dv;
    set_top_level_value(vn,dv);
  } else {
    tuint Nr = 0;
    tuint Nc = 0;
    tdouble* dvp = data->ConstMtxBox.get_Mtx(mfun->eval(),Nr,Nc,false);
    // convert to Octave format
      Matrix rvMtx(Nr,Nc);
      for (tuint i=0;i<Nr;++i) {
	for (tuint j=0;j<Nc;++j) {
	rvMtx(i,j) = *dvp;
	++dvp;
	}
      }
    set_top_level_value(vn,rvMtx);
  }
}

FlxOctave_setvar::~FlxOctave_setvar()
{
  if (vname) delete vname;
  if (sfun) delete sfun;
  if (mfun) delete mfun;
}

FlxObjBase* FlxObjReadOctave_setvar::read()
{
  bool isvec = false;
  {
    const std::string vid = reader->getWord(true,false);
    if (vid=="vector") isvec = true;
    else if (vid=="scalar") isvec = false;
    else {
      std::ostringstream ssV;
      ssV << "Expected either 'scalar' or 'vector' and not '" << vid << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadOctave_setvar::read_1", ssV.str(), reader->getCurrentPos());
    }
  }
  reader->getChar(':');
  FlxString* vname = new FlxString(false,false);
  FlxFunction* sfun = NULL;
  FlxMtxConstFun* mfun = NULL;
  try {
    reader->getChar('=');
    if (isvec) {
      mfun = new FlxMtxConstFun(true);
    } else {
      sfun = new FlxFunction(funReader,false);
    }
    return new FlxOctave_setvar(get_doLog(),vname,sfun,mfun);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadOctave_setvar::read_2",1);
    delete vname;
    if (sfun) delete sfun;
    if (mfun) delete mfun;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

const tdouble FunOctaveGetVar::calc()
{
  return get_top_level_value(varname).scalar_value();
}

const std::string FunOctaveGetVar::write_v()
{
  return "octave_getvar";
}

const std::string FunOctaveGetVar::write()
{
  std::string str1 = write_v() + '(' + varname + ')';
  return str1;
}

FunBase* FunReadFunOctaveGetVar::read( bool errSerious )
{
  // obtain the name of the Octave variable to retrieve
    if ( reader->getNextType() != ReadStream::STRING ) {
      std::ostringstream ssV;
      ssV << "Name of the Octave variable expected.";
      FlxError(errSerious,"FunReadFunOctaveGetVar::read_1", ssV.str(), reader->getCurrentPos());
    }
    const std::string vname = reader->getWord(false,errSerious);
  return new FunOctaveGetVar(vname ); 
}


// ------------------------------------------------------------------------------------------------



